;=============================================================================
;; @name mem.free
;; @desc
;;    Releases allocated block of memory returned from {mem.alloc}, {mem.alloc0},
;;    {mem.realloc}, or {mem.realloc0}.
;; @arg blockptr
;;    Pointer to allocated memory block.
;;    This must be same pointer as returned from allocation function, pointer into block is not enough.
;; @ret
;;    CF set on error
;; @err ERR_INVALID_POINTER
;;    Blockptr is not a valid allocated block pointer.
;;    This could also mean {heap:heap corruption}, these two cases cannot be properly distinguished.
;; @err ERR_HEAP_CORRUPT
;;     {heap:Heap is corrupted}. You was writing to memory outside of allocated block.
;; @err ERR_MODULE_NOT_INITIALIZED
;;      Module is not initialized. Initialize module with {mem.init} to use this procedure.
;; @err ERR_UNKNOWN
;;     System returned error that wasn't translated to FASMLIB error.
;;     You can {contact} author about this, and translation can be added.
;; @note If this procedure returns error, consider block released, and don't use it anymore.
;; @note Memory management must be {mem.init:initialized} for this procedure to work.
;; @note This is platform-specific procedure
proc mem.free blockptr
	push	ecx edx

	;is module initialized?
	cmp	[mem.initialized],0
	jz	.error_not_initialized

	;EAX = memory block (before header)
	mov	eax, [blockptr]
	sub	eax, 8

	;check if it is our memory block
	cmp	dword [eax], 'RnR'
	je	@f
	mov	eax, ERR_INVALID_POINTER
	jmp	.rc
@@:

	;destroy 'RnR' mark - because memory in heap remains accessible even after it is freed
	mov	dword [eax], 0

	;call HeapFree
	push	eax		;pointer to beginning of block (including header)
	push	0		;no flags
	push	[mem.heap]
	call	[HeapFree]
	or	eax,eax
	jz	.error_HeapFree

	;block freed
	dec	[mem.alloc_count]
	jc	.error_alloc_count_overflow

.rnc:	clc
.r:	pop	edx ecx
	ret
.rc:	stc
	jmp	.r

.error_HeapFree:
	mov	eax, ERR_UNKNOWN
	jmp	.rc

.error_not_initialized:
	mov	eax, ERR_MODULE_NOT_INITIALIZED
	jmp	.rc

.error_alloc_count_overflow:
	mov	eax, ERR_FASMLIB
	jmp	.rc
endp