;; @name extconv.dec_iq
;; @desc
;;    Converts decimal string to signed qword (64bit) value.
;;    This is extended version, for simpler version see {conv.dec_iq}.
;; @arg string
;;    Pointer to string to convert.
;; @arg buflen
;;    Size of string's buffer.
;;    You can set this to -1 (FFFFFFFFh) if string is quaranteed to be properly zero-terminated.
;; @arg allowplus
;;    If 1, explicit "+" sign before number is allowed.
;; @arg len
;;    Pointer to dword variable, which will on return hold number of converted characters.
;;    Ignored if zero.
;; @ret
;;    CF=1 on error, otherwise
;;    EDX:EAX = value.
;; @err ERR_INVALID_MODE
;;    allowplus > 1
;; @err ERR_NOT_DECIMAL
;;    First character of string is not decimal digit.
;; @err ERR_OUT_OF_RANGE
;;    Value is greater than 9223372036854775807 or lesser than -9223372036854775808.
;; @err ERR_ZERO_SIZE
;;    strlen=0
;; @note This procedure converts digits until it reaches non-digit character,
;;       or until value overflows.
proc extconv.dec_iq string, strlen, allowplus, len
local sign:DWORD
	push	edi esi ecx ebx

	;check args
	cmp	[strlen], 0
	je	.error_zero_size
	cmp	[allowplus], 1
	ja	.error_mode

	;ESI = string
	mov	esi, [string]

	;ECX = count of converted chars
	mov	ecx, 0

	;get sign
	mov	[sign], 0
	cmp	byte [esi], '-'
	jne	taken @f
	inc	esi
	inc	ecx
	mov	[sign], 1
	jmp	.sign_done
	@@:
	cmp	[allowplus], 0
	je	.sign_done
	cmp	byte [esi], '+'
	jne	.sign_done
	inc	esi
	inc	ecx
.sign_done:

	;error if first char isn't digit
	cmp	byte [esi], '0'
	jb	.error_not_dec
	cmp	byte [esi], '9'
	ja	.error_not_dec

	;EDI:EBX = final value, for now 0
	mov	edi, 0
	mov	ebx, 0

	;loop
.digit_loop:
	;AL=char
	mov	al, [esi]
	inc	esi

	;end if it isn't hex digit
	cmp	al, '0'
	jb	.done
	cmp	al, '9'
	ja	.done

	;convert to binary digit
	sub	al, '0'
	movzx	eax, al

	;multiply value with 10
	push	eax esi
	xor	edx, edx
	mov	esi, 10
	mov	eax, ebx
	mul	esi
	mov	ebx, eax
	mov	eax, edi
	mov	edi, edx
	mul	esi
	add	eax, edi
	mov	edi, eax
	pop	esi eax

	;add digit
	add	ebx, eax
	adc	edi, 0

	;check for overflow
	jc	.overflow
	cmp	edx, 0
	ja	.overflow

	;continue loop
	inc	ecx
	cmp	ecx, [strlen]
	jb	.digit_loop
.done:

	;fill len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	;set sign and check overflow
	cmp	[sign], 0
	jne	.negative
.positive:
	cmp	edi, 0x7FFFFFFF
	ja	untaken .overflow
	jmp	.sign_set
.negative:
	cmp	edi, 0x80000000
	jb	taken .negative_okay
	ja	untaken .overflow
	cmp	ebx, 0
	jne	untaken .overflow
	jmp	.sign_set  ;number is exactly 0x8000000000000000
.negative_okay:
	not	ebx
	not	edi
	add	ebx, 1
	adc	edi, 0
.sign_set:

	mov	edx, edi
	mov	eax, ebx


	clc
.r:	pop	ebx ecx esi edi
	ret
.rc:	stc
	jmp	.r


.overflow:
	;set len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], ecx
	@@:

	;return error
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc

.error_zero_size:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	;return error
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_mode:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	;return error
	mov	eax, ERR_INVALID_MODE
	jmp	.rc

.error_not_dec:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	;return error
	mov	eax, ERR_NOT_DECIMAL
	jmp	.rc
endp