;; @name fileio.append
;; @desc
;;    Opens existing file for writing, and sets pointer to end of file.
;; @arg name
;;    Pointer to name of file to open. Must be zero-terminated string.
;; @arg create
;;    Specifies whether to create file if it doesn't exist.
;;    1=create, 0=don't create.
;; @arg mode
;;    Access mode to file: 1=write only, 2=read/write.
;; @ret
;;    CF = 1 on error, otherwise
;;    EAX = file handle
;; @err ERR_FILE_NOT_FOUND
;;    File with specified name wasn't found.
;;    Only returned when create=0.
;; @err ERR_CANT_OPEN_FILE
;;    File exists, but couldn't be opened.
;; @err ERR_INVALID_MODE
;;    mode=0, or mode>2, or create>1.
;; @err ERR_UNKNOWN
;;    Platform-specific error during opening file.
;; @warn Don't forget to close returned handle with {fileio.close}!
;; @note This is platform-dependent procedure
proc fileio.append name, create, mode
	push	ebx

	;check if mode is 1 or 2
	mov	eax, [mode]
	test	eax, eax
	jz	.error_invalid_mode
	cmp	eax, 2
	ja	.error_invalid_mode

	;check if create is 0 or 1
	cmp	[create], 1
	ja	.error_invalid_mode

	;try to open file
	libcall fileio.open, [name], [mode]

	;check if file was opened successfully
	jnc	.file_opened

	;file doesn't exist?
	cmp	eax, ERR_FILE_NOT_FOUND
	jne	.rc

	;should we create file?
	cmp	[create], 1
	jne	.rc

	;create file
	libcall fileio.create, [name], 1, [mode]
	jc	.rc

.file_opened:

	;EBX = file handle
	mov	ebx, eax

	;now seek end of file
	libcall fileio.seek, ebx, 0, 0, 2
	jc	.rc2

	;okay, return handle
	mov	eax, ebx

.rnc:	clc
.r:	pop	ebx
	ret
.rc:	stc
	jmp	.r

;try to close EBX file and return error
.rc2:	push	eax
	libcall fileio.close, ebx
	pop	eax
	jmp	.rc

.error_invalid_mode:
	mov	eax, ERR_INVALID_MODE
	jmp	.rc

endp