;============================================================================
;; @name mem.cmp
;; @desc
;;    Compares contents of two blocks of memory.
;;    If they are different, returns offset of position where they are different.
;;    Additionally it says which block contents "is bigger", useful for sorting.
;; @arg block1
;;    Pointer to beginning of first memory block
;; @arg block2
;;    Pointer to beginning of second memory block
;; @arg size
;;    Number of bytes to compare
;; @ret
;;    CF = 1 on error, otherwise
;;    {flagscmp:flags set for signed conditional jumps}, and
;;    if ZF=0, EAX = offset in blocks (not pointer) where blocks differ
;; @err ERR_INVALID_POINTER
;;    some of pointers point to inaccessible area
;; @err ERR_ZERO_SIZE
;;    size=0
;; @note   When ZF=1 (blocks are equal), EAX=size on return
;; @note   Memory management doesn't have to be initalized for this to work
proc mem.cmp block1, block2, size
	push	esi edi ecx
	pushf
	cld

	mov	edi, [block1]
	mov	esi, [block2]
	mov	ecx, [size]

	;error if size=0
	cmp	ecx, 0
	je	.error_size_zero

	;do the compare
	repe	cmpsb
	je	.equal

	;set flags to greater/lesser
	mov	al, byte [edi-1]
	cmp	al, byte [esi-1]
	ja	.greater

	;flags for conditional jumps
	; je:      ZF=1
	; jne:     ZF=0
	; jg/jnle: ZF=0 & (SF=OF)
	; jge/jnl: SF=OF
	; jl/jnge: SF!=OF
	; jle/jng: ZF=1 | (SF!=OF)
.lesser:
	;ZF=0, SF=0, OF=1
	and	word [esp], not ((1 shl 6) + (1 shl 7)) ;clear ZF and SF
	or	word [esp], 1 shl 11			;set OF
	jmp	.done

.greater:
	;ZF=0, SF=1, OF=1
	and	word [esp], not (1 shl 6)		;clear ZF
	or	word [esp], (1 shl 11) + (1 shl 7)	;set OF and SF
	jmp	.done

.equal:
	;ZF=1, SF=1, OF=1
	or	word [esp], (1 shl 6) + (1 shl 7) + (1 shl 11)

.done:	;EAX = offset where blocks differ
	mov	eax, [size]
	sub	eax, ecx
	dec	eax

.rnc:	and	byte [esp],not 1     ;clear CF
.r:	popf
	pop	ecx edi esi
	ret
.rc:	or	byte [esp], 1 ;set CF
	jmp	.r

.error_size_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc
endp