;; @name str.ins
;; @desc
;;    Inserts string to into another string.
;; @arg dest
;;    Pointer to destination string, the one into which we insert.
;; @arg destlen
;;    Size of destination string buffer.
;; @arg src
;;    Pointer to source string, the one we are inserting.
;; @arg srclen
;;    Size of source string buffer.
;;    You can set this to -1 (FFFFFFFFh) if string is quaranteed to be properly zero-terminated.
;; @arg pos
;;    Position in destination string where we are insrting.
;;    Position 0 corresponds to beginning of string, before first character.
;; @ret
;;    CF set on error, otherwise
;;    EAX = dest.
;; @err ERR_OUT_OF_RANGE
;;    Position is outside of destination string, pos > strlen(dest).
;;    You are trying to insert behind end of string.
;; @err ERR_BUFFER_OVERFLOW
;;    Destination buffer is not big enough.
;;    Only first destlen bytes of resulting string were written.
;;    String in destination buffer is NOT zero terminated.
;; @err ERR_ZERO_SIZE
;;    destlen=0 or srclen=0.
proc str.ins dest, destlen, src, srclen, pos
	push	ebx ecx esi edi
	cld

	;error if destlen=0
	cmp	[destlen], 0
	je	.error_destlen_zero

	;error if srclen=0
	cmp	[srclen], 0
	je	.error_srclen_zero

	;ECX = length of source string
	libcall str.len, [src], [srclen]
	jc	.rc
	mov	ecx, eax

	;just return when source string is empty
	;??? - returns without ERR_BUFFER_FULL even when dest string is not terminated...
	cmp	ecx, 0
	je	.done

	;EBX = length of dest string, return with OF if not zero terminated
	libcall str.len, [dest], [destlen]
	jc	.rc
	mov	ebx, eax

	;check if pos is inside dest string (pos <= str.len(dest))
	cmp	[pos], ebx
	ja	.error_position

	;we don't need to shift characters from dest string if source string
	;fills entire buffer, eg. str.len(src) >= destlen-pos
	mov	eax, [destlen]
	sub	eax, [pos]
	cmp	ecx, eax
	ja	.copy_src

	;we also don't need to shift characters from dest string if we are
	;inserting to it's end
	cmp	[pos], ebx
	je	.copy_src

	;but we have to check case when there is not enough space for moving
	;all of characters, eg str.len(dest) + str.len(src) > destlen
	lea	eax, [ebx+ecx]
	cmp	eax, [destlen]
	jbe	.can_shift_all

	;we can't shift all characters. find how many we can
	;it's: destlen - (pos + str.len(src))
	mov	eax, [destlen]
	sub	eax, [pos]
	sub	eax, ecx
	jmp	.shift_dest

	;we can shift all characters - that means str.len(dest)-pos
.can_shift_all:
	mov	eax, ebx
	sub	eax, [pos]

	;move EAX characters by ECX places forward to make space for new string
.shift_dest:
	mov	esi, [dest]
	add	esi, [pos]    ;from dest[pos]
	mov	edi, esi
	add	edi, ecx      ;to dest[pos+str.len(src)]
	libcall mem.copy, edi, esi, eax
	jc	.rc

	;copy source into destination
.copy_src:

	;EAX = how many bytes we can copy = min(str.len(src), destlen-pos)
	mov	eax, [destlen]
	sub	eax, [pos]
	cmp	eax, ecx
	cmova	eax, ecx

	;error if we are trying to insert at end of string which is also end of buffer
	cmp	eax, 0
	je	.error_dest_buffer_full

	;let us copy
	mov	edi, [dest]
	add	edi, [pos]
	libcall mem.copy, edi, [src], eax

	;error if there isn't space left for ending zero
	lea	eax, [ebx+ecx]
	cmp	eax, [destlen]
	jae	.error_dest_buffer_full

	;now add ending zero to end of string, eg. to: min(str.len(dest)+str.len(src), destlen-1)
	lea	eax, [ebx+ecx]
	mov	ebx, [destlen]
	dec	ebx
	cmp	eax, ebx
	cmova	eax, ebx
	add	eax, [dest]
	mov	byte [eax], 0


.done:	mov	eax,[dest]
	clc
.r:	pop	edi esi ecx ebx
	ret

.rc:	stc
	jmp	.r

.error_dest_buffer_full:
	mov	eax, ERR_BUFFER_FULL
	jmp	.rc

.error_destlen_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_srclen_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_position:
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc
endp
