;; @name stream._def_peek
;; @desc
;;    This is default handler for peek method. It uses callback procedure
;;    to read data into stream read buffer, and returns pointer to it.
;; @arg stream
;;    Handle of stream.
;; @arg count
;;    Number of bytes to peek.
;; @arg readlen
;;    Pointer to variable, which will hold number of chars in returned buffer.
;;    Only needed when function returns OF, otherwise it's equal to count.
;;    You can set this to 0, when you don't want handle case with OF=1.
;; @arg readbuf
;;    Pointer to procedure that does actual reading. Procedure
;;    recieves pointer to memory where to read, and number of bytes to read.
;;    On return, EAX holds number of bytes readen.
;; @ret
;;    CF set on error, otherwise
;;    OF set if EOF was reached (less than count bytes readen), and
;;    EAX = pointer to buffer (READ ONLY, do not write there)
;; @warn Pointer to buffer becomes invalid after next operation with stream.
;; @warn On OF=1, do not read buffer behind number of bytes returned in readlen variable.
;;       If you passed 0 as readlen, don't read from buffer at all.
;; @note Stream must have peek method supported, or sufficently large read buffer.
;; @err ERR_ZERO_SIZE
;;    Count=0
;; @err ERR_INVALID_HANDLE
;;    Not a stream handle.
;; @err ERR_HANDLE_NO_BUFFER
;;    Stream doesn't have read buffer associated, nor peek method supported.
;; @err ERR_HANDLE_NOT_READABLE
;;    Stream is not readable.
;; @err ERR_BUFFER_FULL
;;    Handle is associated with static buffer, with length smaller than count.
;;    Nothing was read from stream.
;; @err ERR_HANDLE_CORRUPT
;;    Stream is corrupt, you are overwriting stream object somewhere
;;    (or it's error in FASMLIB).
proc stream._def_peek, stream, count, readlen, readbuf
	push	ebx ecx edi
	pushf

	;EBX = stream object
	mov	ebx, [stream]

	;check if buffer type is valid
	cmp	[ebx+STREAM.rbuf.type], 2
	ja	.error_corrupt_stream

	;error if count is zero
	cmp	[count], 0
	je	.error_count_zero

	;error if stream is not readable
	mov	eax, [ebx + STREAM.vtab]
	cmp	[eax + STREAM_VTAB.read], 0
	je	.error_stream_not_readable

	;error if cnt > size-pos
	mov	eax, [ebx+STREAM.rbuf.size]
	sub	eax, [ebx+STREAM.rbuf.pos]
	cmp	eax, [ebx+STREAM.rbuf.cnt]
	jb	.error_corrupt_stream

	;ECX = count
	mov	ecx, [count]

	;error if we doesn't have any buffer
	cmp	[ebx+STREAM.rbuf.type], 0
	je	.error_no_buffer

	;error if we have static buffer smaller than count
	cmp	[ebx+STREAM.rbuf.type], 1
	jne	@f
	cmp	ecx, [ebx+STREAM.rbuf.size]
	ja	.error_buffer_small
	@@:

	;if we have dynamic buffer, not yet allocated, then allocate it
	cmp	[ebx+STREAM.rbuf.type], 2
	jne	@f
	cmp	[ebx+STREAM.rbuf], 0
	jne	@f
	libcall mem.alloc, ecx
	jc	.rc
	mov	[ebx+STREAM.rbuf], eax
	mov	[ebx+STREAM.rbuf.size], ecx
	mov	[ebx+STREAM.rbuf.pos], 0
	mov	[ebx+STREAM.rbuf.cnt], 0
	jmp	.buffer_long_enough
	@@:

	;see if buffer is big enough for count (without moving needed)
	mov	eax, [ebx+STREAM.rbuf.size]
	sub	eax, [ebx+STREAM.rbuf.pos]
	cmp	ecx, eax
	jbe	.buffer_long_enough

	;move data to beginning of buffer, if they aren't
	cmp	[ebx+STREAM.rbuf.pos], 0
	ja	@f
	mov	eax, [ebx+STREAM.rbuf.pos]
	add	eax, [ebx+STREAM.rbuf]
	libcall mem.copy, [ebx+STREAM.rbuf], eax, [ebx+STREAM.rbuf.cnt]
	jc	.rc
	mov	[ebx+STREAM.rbuf.pos], 0
	@@:

	;see if buffer is long enough now
	cmp	ecx, [ebx+STREAM.rbuf.size]
	jbe	.buffer_long_enough

	;if not, reallocate buffer to needed size
	libcall mem.realloc, [ebx+STREAM.rbuf], ecx
	jc	.rc
	mov	[ebx+STREAM.rbuf], eax
	mov	[ebx+STREAM.rbuf.size], ecx

.buffer_long_enough:

	;check if there is already enough bytes in buffer
	cmp	ecx, [ebx+STREAM.rbuf.cnt]
	jbe	.buffer_filled

	;check if there is EOF marked in buffer
	cmp	[ebx+STREAM.rbuf.eof],1
	je	.buffer_filled

	;read remaining bytes
	mov	edi, [ebx+STREAM.rbuf]
	add	edi, [ebx+STREAM.rbuf.pos]
	add	edi, [ebx+STREAM.rbuf.cnt]
	sub	ecx, [ebx+STREAM.rbuf.cnt]
	libcall [readbuf], ebx, edi, ecx
	jc	.rc

	;adjust buffer info (count, EOF flag)
	cmp	eax, ecx
	;TODO - return some error if EAX > ECX
	je	@f
	mov	[ebx+STREAM.rbuf.eof], 1
	@@:
	add	[ebx+STREAM.rbuf.cnt], eax

.buffer_filled:

	;readlen = min(count, rbuf.cnt)
	mov	ecx, [count]
	mov	eax, [ebx+STREAM.rbuf.cnt]
	cmp	ecx, eax
	cmova	ecx, eax
	mov	eax, [readlen]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	;eax = rbuf
	mov	eax, [ebx + STREAM.rbuf]
	add	eax, [ebx + STREAM.rbuf.pos]

	;if user didn't set readlen, and we are returning OF, then return 0 in EAX,
	;so he immediately get GPF if he tries to read from buffer.
	cmp	[readlen], 0
	jne	@f
	cmp	ecx, [count]
	jnb	@f
	mov	eax, 0
	@@:

	;return OF if count<rbuf.cnt
	cmp	ecx, [count]
	jb	.ro

.rno:	and	word [esp], not (1 shl 11) ;clear OF
.rnc:	and	byte [esp], not 1 ;clear CF
.r:	popf
	pop	edi ecx ebx
	ret
.rc:	or	byte [esp], 1 ;set CF
	jmp	.r
.ro:	or	word [esp], 1 shl 11 ;set OF
	jmp	.rnc

.error_count_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_not_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc

.error_corrupt_stream:
	mov	eax, ERR_HANDLE_CORRUPT
	jmp	.rc

.error_no_buffer:
	mov	eax, ERR_HANDLE_NO_BUFFER
	jmp	.rc

.error_buffer_small:
	mov	eax, ERR_BUFFER_FULL
	jmp	.rc

.error_stream_not_readable:
	mov	eax, ERR_HANDLE_NOT_READABLE
	jmp	.rc
endp