;; @name stream._def_read
;; @desc
;;     This is internal procedure, used by {stream read method|stream read methods}.
;;     It needs pointer to callback procedure, which performs actual
;;     reading of data to given place in memory.
;;     This procedure takes care of maintaing stream read buffer for any type of stream.
;; @arg stream
;;    Handle of stream from which to read.
;; @arg buffer
;;    Pointer to buffer where data will be stored.
;; @arg count
;;    Number of bytes to read. Cannot be 0.
;; @arg readbuf
;;    Pointer to procedure that does actual reading. Procedure
;;    recieves pointer to memory where to read, and number of bytes to read.
;;    On return, EAX holds number of bytes readen.
;; @ret
;;    CF set on error, otherwise
;;    OF set if EOF was reached, and
;;    EAX = number of bytes readen.
;;    If OF=0, then EAX=count.
;; @err ERR_INVALID_HANDLE
;;    Invalid stream handle.
;; @err ERR_HANDLE_NOT_READABLE
;;    Stream is not readable.
;; @err ERR_ZERO_SIZE
;;    count=0.
;; @err ERR_HANDLE_CORRUPT
;;    Stream is corrupt.
;;    Either you are overwriting stream object somewhere, or it is an error in FASMLIB.
;; @dep +readbuf
;;    Errors returned by callback procedure are returned to caller.
proc stream._def_read stream, buffer, count, readbuf
	push	ebx ecx edx edi
	pushf

	;EBX = stream
	mov	ebx, [stream]

	;check if buffer type is valid
	cmp	[ebx+STREAM.rbuf.type], 2
	ja	.error_corrupt_stream

	;error if cnt > size-pos
	mov	eax, [ebx+STREAM.rbuf.size]
	sub	eax, [ebx+STREAM.rbuf.pos]
	cmp	eax, [ebx+STREAM.rbuf.cnt]
	jb	.error_corrupt_stream

	;EDI = buffer, ECX = count
	mov	edi, [buffer]
	mov	ecx, [count]

	;if there is no buffering, go directly to read method
	cmp	[ebx+STREAM.rbuf.type], 0
	je	.call_read

	;check if there is some data on buffer
	cmp	[ebx+STREAM.rbuf.cnt], 0
	ja	.read_buffer

	;no data on buffer, so if EOF flag is clear, call read method
	cmp	[ebx+STREAM.rbuf.eof], 0
	je	.call_read

	;no data on buffer, EOF flag set, return EAX=0 and OF=1
	mov	eax, 0
	jmp	.ro

.read_buffer:

	;EDX = bytes to copy from stream buffer = min(count, stream.rbuf.cnt)
	mov	edx, [ebx+STREAM.rbuf.cnt]
	cmp	edx, ecx
	cmova	edx, ecx

	;copy from stream buffer to dest buffer
	mov	eax, [ebx+STREAM.rbuf]
	add	eax, [ebx+STREAM.rbuf.pos]
	libcall mem.copy, edi, eax, edx
	jc	.rc

	;remove data from buffer (set buffer pointer behind them)
	add	[ebx+STREAM.rbuf.pos], edx
	sub	[ebx+STREAM.rbuf.cnt], edx
	jnz	@f
	mov	[ebx+STREAM.rbuf.pos], 0   ;if we emptied the stream buffer, set pointer to beginning
	@@:

	;eax = number of bytes readen
	mov	eax, edx

	;set new destination buffer's position and length
	add	edi, edx
	sub	ecx, edx

	;if we have read enough, return EAX=count, OF=0
	jnz	@f
	mov	eax, [count]
	jmp	.rno
	@@:

	;if EOF flag is set in stream buffer, return EAX=readen, OF=1
	cmp	[ebx + STREAM.rbuf.eof], 1
	jne	@f
	mov	eax, edx
	jmp	.ro
	@@:

	;if we didn't read enough, continue reading with read method

.call_read:

	;call read method
	libcall [readbuf], ebx, [buffer], ecx
	jc	.rc

	;store size readen. if OF was returned, then mark it on buffer
	cmp	eax, ecx
	;TODO: error if returned EAX > ECX
	je	@f
	mov	[ebx+STREAM.rbuf.eof], 1
	@@:
	sub	ecx, eax

.done:

	;EAX = size readen from buffer = count-ECX
	mov	eax, [count]
	sub	eax, ecx

	;return OF if EOF was reached
	cmp	[ebx+STREAM.rbuf.eof], 1
	je	.ro

.rno:	and	word [esp], not (1 shl 11) ;clear OF
.rnc:	and	byte [esp], not 1 ;clear CF
.r:	popf
	pop	edi edx ecx ebx
	ret
.rc:	or	byte [esp], 1 ;set CF
	jmp	.r
.ro:	or	word [esp], 1 shl 11 ;set OF
	jmp	.rnc

.error_not_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc

.error_stream_not_readable:
	mov	eax, ERR_HANDLE_NOT_READABLE
	jmp	.rc

.error_count_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_corrupt_stream:
	mov	eax, ERR_HANDLE_CORRUPT
	jmp	.rc
endp