;; @name stream.close
;; @desc
;;    Closes stream handle. If stream object was allocated,
;;    or it's buffers were allocated, they are deallocated.
;; @arg stream
;;    Stream handle to close.
;; @ret
;;    CF set on error.
;; @err ERR_INVALID_HANDLE
;;    Not a valid stream handle.
;; @err ERR_HANDLE_CORRUPT
;;    Handle is corrupt. Probably you have overwriten stream
;;    object in memory. This might also be a FASMLIB bug.
;; @dep {stream close method}
proc stream.close, stream
	push	ebx

	;EBX = stream object, EAX = virtual table
	mov	ebx, [stream]
	mov	eax, [ebx + STREAM.vtab]

	;error if invalid stream
	cmp	[eax + STREAM_VTAB.sign], 'STRM'
	jne	.error_not_stream

	;EAX = close method, skip if not present
	mov	eax, [eax + STREAM_VTAB.close]
	cmp	eax, 0
	je	@f
	libcall eax, [stream]
	jc	.error_method_failed
	@@:

	;destroy read buffer if it is dynamic and allocated
	cmp	[ebx+STREAM.rbuf.type], 2
	jne	@f
	cmp	[ebx+STREAM.rbuf], 0
	je	@f
	libcall mem.free, [ebx+STREAM.rbuf]
	jc	.error_buffer_freeing_failed
	@@:

	;destroy write buffer if it is dynamic and allocated
	cmp	[ebx+STREAM.wbuf.type], 2
	jne	@f
	cmp	[ebx+STREAM.wbuf], 0
	je	@f
	libcall mem.free, [ebx+STREAM.wbuf]
	jc	.error_buffer_freeing_failed
	@@:

	;delete pointer to vtable (safety)
	mov	[ebx+STREAM.vtab], 0

	;destroy object
	cmp	[ebx+STREAM.allocated], 1
	jne	@f
	libcall mem.free, ebx
	jc	.rc
	@@:


	clc
.r:	pop	ebx
	ret
.rc:	stc
	jmp	.r

.error_not_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc

.error_method_failed:
	push	eax
	libcall mem.free, [stream]
	pop	eax
	jmp	.rc

.error_buffer_freeing_failed:
	push	eax
	libcall mem.free, [stream]
	pop	eax

	;translate ERR_INVALID_POINTER to ERR_HANDLE_CORRUPT (pointer to buffer was invalid)
	cmp	eax, ERR_INVALID_POINTER
	jne	@f
	mov	eax, ERR_HANDLE_CORRUPT
	jmp	.rc
	@@:

	jmp	.rc

endp