;; @name stream.create
;; @desc
;;    Creates stream object.
;;    This call allows you to create static objects, dynamic object with
;;    static buffer, or dynamic objects with dynamic buffer.
;; @arg objbuf
;;    Pointer to buffer to hold stream object. size of buffer must be at least "sizeof.STREAM".
;;    Set to 0 if you want procedure to allocate stream.
;; @arg size
;;    Number of additional bytes to allocate with object.
;;    These can be used to store additional data for methods.
;;    This value must be 0, if objbuf is nonzero (static stream object).
;; @arg vtab
;;    Points to object's {stream virtual table:virtual table}.
;;    If readbuf=0, then {stream peek method:peek method} in virtual table must be 0.
;; @arg readbuf
;;    Pointer to read buffer.
;;    0 if none, 1 for dynamic buffer.
;;    Must be 0 if read method pointer is 0 in vtab.
;; @arg rbuflen
;;    Size of read buffer.
;;    If readbuf=0, ignored.
;;    If readbuf=1, initial size of read buffer (allocated at creation).
;;    If readbuf=pointer, size of buffer.
;;    Must be 0 if read method pointer is 0 in vtab.
;; @arg writebuf
;;    Pointer to write buffer, 0 if none, 1 for dynamic buffer.
;;    Must be 0 if write method pointer is 0 in vtab.
;; @arg wbuflen
;;    Size of write buffer.
;;    If writebuf=0, ignored.
;;    If writebuf=1, initial size of buffer (allocated at creation).
;;    If writebuf=pointer, size of buffer.
;;    Must be 0 if write method pointer is 0 in vtab.
;; @ret
;;    CF=1 on error, otherwise
;;    EAX = pointer to created stream object (eg. handle of stream).
;; @warn Don't forget to close stream with {stream.close}
;; @warn If you disable read buffering (rbuf=0), then you can't use {stream.peek:peek},
;;       or any procedures that require peek.
;;       Most {text read:text reading} procedures will fail.
;; @note When pointer=0 and size>0, then additional memory is filled with zeroes.
;; @note To create default (completely dynamic) stream,
;;       set objbuf=0, readbuf=1, rbuflen=0, writebuf=1, wbuflen=0.
;; @err ERR_HANDLE_NOT_READABLE
;;    read=0 and some of peek, skip, readbug, rbuflen is non-zero.
;; @err ERR_OUT_OF_RANGE
;;    size<0.
;; @dep mem.alloc0
;;    When pointer=0.
;; @dep mem.alloc
;;    When readbuf=1 or writebuf=1.
proc stream.create	objbuf, size, vtab, readbuf, rbuflen, writebuf, wbuflen
	push	ebx

	;if objbuf!=0, then size must be 0
	cmp	[objbuf], 0
	je	@f
	cmp	[size], 0
	jne	.error_arguments
	@@:

	;if readbuf=0, then peek, peeknb and rbuflen must be 0
	cmp	[readbuf], 0
	jne	@f
	cmp	[rbuflen], 0
	jne	.error_arguments
	@@:

	;TODO check methods in vtab

	;catch negative additional size
	cmp	[size], 0
	jl	.error_negative_size

	;EBX = pointer to stream object buffer, allocate it if requested
	mov	ebx, [objbuf]
	cmp	ebx, 0
	jne	@f
	mov	eax, [size]
	add	eax, sizeof.STREAM
	libcall mem.alloc, eax
	jc	.rc
	mov	ebx, eax
	@@:

	;zero stream object
	libcall mem.fill, ebx, sizeof.STREAM, 0
	jc	.rc

	;set vtab
	mov	eax, [vtab]
	mov	[ebx+STREAM.vtab], eax

	;set "allocated" flag
	mov	eax, [objbuf]
	cmp	eax, 0
	jne	@f
	mov	[ebx+STREAM.allocated], 1
	@@:

	;fill read buffer info, if read method is supported
	mov	eax, [vtab]
	cmp	[eax + STREAM_VTAB.read], 0
	je	.rbuf_done

	;check if read buffering is disabled
	cmp	[readbuf], 0
	je	.rbuf_done

	;if it's static buffer, just set values
	cmp	[readbuf], 1
	je	@f
	mov	[ebx+STREAM.rbuf.type], 1
	mov	eax, [readbuf]
	mov	[ebx+STREAM.rbuf], eax
	mov	eax, [rbuflen]
	mov	[ebx+STREAM.rbuf.size], eax
	jmp	.rbuf_done
	@@:

	;it's dynamic buffer with nonzero initial size, allocate it
	mov	[ebx+STREAM.rbuf.type], 2
	cmp	[rbuflen], 0
	je	@f
	libcall mem.alloc, [rbuflen]
	jc	.rc2
	mov	[ebx + STREAM.rbuf], eax
	mov	eax, [rbuflen]
	mov	[ebx+STREAM.rbuf.size], eax
	@@:

.rbuf_done:

	;check if write buffering is disabled
	cmp	[writebuf], 0
	je	.wbuf_done

	;if it's static buffer, just set values
	cmp	[writebuf], 1
	je	@f
	mov	[ebx+STREAM.wbuf.type], 1
	mov	eax, [writebuf]
	mov	[ebx+STREAM.wbuf], eax
	mov	eax, [wbuflen]
	mov	[ebx+STREAM.wbuf.size], eax
	jmp	.wbuf_done
	@@:

	;it's dynamic buffer with nonzero initial size, allocate it
	mov	[ebx+STREAM.wbuf.type], 2
	cmp	[wbuflen], 0
	je	.wbuf_done
	libcall mem.alloc, [wbuflen]
	jc	.rc2
	mov	[ebx + STREAM.wbuf], eax
	mov	eax, [wbuflen]
	mov	[ebx+STREAM.wbuf.size], eax

.wbuf_done:

	;return pointer to stream
	mov	eax, ebx

.r:	pop	ebx
	ret
.rc:	stc
	jmp	.r

.rc2:	push	eax
	cmp	[ebx+STREAM.rbuf], 0
	je	@f
	libcall mem.free, [ebx+STREAM.rbuf]
	@@:
	cmp	[ebx+STREAM.wbuf], 0
	je	@f
	libcall mem.free, [ebx+STREAM.wbuf]
	@@:
	libcall mem.free, ebx
	pop	eax
	jmp	.rc

.error_negative_size:
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc

.error_arguments:
	mov	eax, ERR_HANDLE_NOT_READABLE
	jmp	.rc
endp