;; @name stream.peek
;; @desc
;;    Pre-reads data from stream, and returns you pointer to them.
;;    Data still remain in stream, and will be readen by subsequent reads.
;;    Returned pointer is only valid until next operation with stream.
;;    Blocks until there is enough data available.
;; @arg stream
;;    Handle of stream.
;; @arg count
;;    Minimal number of bytes to peek.
;; @arg readlen
;;    Pointer to variable, which will hold number of chars in returned buffer.
;;    When function returns OF=0, this number is greater or equal to count.
;;    When function returns OF=1, this number is less than than count.
;;    You can set this pointer to 0.
;; @ret
;;    CF set on error, otherwise
;;    OF set if EOF was reached (less than count bytes readen), and
;;    EAX = pointer to buffer (READ ONLY, do not write there)
;; @warn Pointer to buffer becomes invalid after next operation with stream.
;; @warn On OF=1, do not read buffer behind number of bytes returned in readlen variable.
;;       If you passed 0 as readlen, don't read from buffer at all.
;; @note Stream must have peek method supported, or sufficently large read buffer.
;; @err ERR_ZERO_SIZE
;;    Count=0
;; @err ERR_INVALID_HANDLE
;;    Not a stream handle.
;; @err ERR_HANDLE_NO_BUFFER
;;    Stream doesn't have read buffer associated, nor peek method supported.
;; @err ERR_HANDLE_NOT_PEEKABLE
;;    Stream is not peekable.
;; @err ERR_BUFFER_FULL
;;    Handle is associated with static buffer, with length smaller than count.
;;    Nothing was read from stream.
;; @err ERR_HANDLE_CORRUPT
;;    Stream is corrupt, you are overwriting stream object somewhere
;;    (or it's error in FASMLIB).
;; @dep {stream peek method}
proc stream.peek, stream, count, readlen

	;do nothing if count is zero
	cmp	[count], 0
	je	.rnc

	;EAX = stream virtual table
	mov	eax, [stream]
	mov	eax, [eax + STREAM.vtab]

	;check if this is proper stream
	cmp	[eax + STREAM_VTAB.sign], 'STRM'
	jne	.error_not_stream

	;EAX = peek method, error if not present
	mov	eax, [eax + STREAM_VTAB.peek]
	cmp	eax, 0
	je	.error_stream_not_peekable

	;call peek method
	libcall eax, [stream], [count], [readlen]
	jc	.rc

.rnc:	clc
.r:	ret
.rc:	stc
	jmp	.r

.error_not_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc

.error_stream_not_peekable:
	mov	eax, ERR_HANDLE_NOT_PEEKABLE
	jmp	.rc


endp