;; @name stream.skip
;; @desc
;;    Skips data from stream input.
;;    Skipping is like reading from stream, without actually storing readen data anywhere.
;; @arg stream
;;    Handle of stream to skip from.
;; @arg count
;;    Number of bytes to skip. Procedure does nothing if this is 0.
;; @ret
;;    CF set on error, otherwise
;;    OF set if EOF was reached, and
;;    EAX = number of bytes skipped.
;;    If OF=0, then EAX=count.
;; @err ERR_INVALID_HANDLE
;;    Invalid stream handle passed.
;; @err ERR_HANDLE_NOT_READABLE
;;    Stream is not readable.
;; @warn
;;    Don't use skip method to seek forward long distances.
;;    There is {stream.seek} for that.
;;    Skipping should be only used for small amount of data,
;;    and on streams that don't support seeking.
;; @dep {stream skip method}
;; @dep {stream read method}
;;    Used when stream doesn't have a skip method.
proc stream.skip stream, count
	push	ebx ecx edx edi
	pushf

	;EBX = stream object, error if invalid
	mov	ebx, [stream]
	mov	eax, [ebx + STREAM.vtab]
	cmp	[eax+STREAM_VTAB.sign], 'STRM'
	jne	.error_not_stream

	;check if buffer type is valid
	cmp	[ebx+STREAM.rbuf.type], 2
	ja	.error_corrupt_stream

	;error if cnt > size-pos
	mov	eax, [ebx+STREAM.rbuf.size]
	sub	eax, [ebx+STREAM.rbuf.pos]
	cmp	eax, [ebx+STREAM.rbuf.cnt]
	jb	.error_corrupt_stream

	;ECX = bytes to skip
	mov	ecx, [count]

	;if it's zero, just return
	cmp	ecx, 0
	cmove	eax, ecx
	je	.rno

	;if we don't have buffering enabled, go directly to method call
	cmp	[ebx+STREAM.rbuf.type], 0
	je	.call_method

	;if we don't have anything on buffer, go to method call
	cmp	[ebx+STREAM.rbuf.cnt], 0
	je	.call_method

	;EDX = bytes to remove from buffer = min(count, bytes_on_buffer)
	mov	edx, ecx
	mov	eax, [ebx+STREAM.rbuf.cnt]
	cmp	edx, eax
	cmova	edx, eax

	;remove them from buffer
	add	[ebx+STREAM.rbuf.pos], edx
	sub	[ebx+STREAM.rbuf.cnt], edx
	jnz	@f
	mov	[ebx+STREAM.rbuf.pos], 0 ;we cleared entire buffer, so move pos to beginning
	@@:

	;end, if enough data removed
	sub	ecx, edx
	jne	@f
	mov	eax, [count]
	jmp	.rno
	@@:

	;end, if EOF was stored on buffer
	cmp	[ebx+STREAM.rbuf.eof], 1
	jne	@f
	mov	eax, [count]
	sub	eax, ecx
	jmp	.ro
	@@:

	;otherwise, continue removing rest with skip or read

.call_method:

	;eax = skip method, use read if unsupported
	mov	eax, [ebx + STREAM.vtab]
	mov	eax, [eax + STREAM_VTAB.skip]
	cmp	eax, 0
	je	.skip_method_unsupported
	libcall eax, [stream], ecx

	;done if we skipped enough bytes without reaching EOF
	jo	@f
	mov	eax, [count]
	jmp	.rno
	@@:

	;if we reached eof, set EOF flag and return number of bytes skipped
	mov	[ebx+STREAM.rbuf.eof], 1
	sub	ecx, eax     ;ecx = bytes_to_skip (except those skipped from buffer) - bytes_skipped
	mov	eax, [count]
	sub	eax, ecx     ;eax = bytes_to_skip_total - ecx
	jmp	.ro

.skip_method_unsupported:

	;last resort - we have to use read method
	;to read data into dummy buffer
.use_read_method:

	;error if stream is not readable
	mov	eax, [ebx + STREAM.vtab]
	cmp	[eax + STREAM_VTAB.read], 0
	je	.error_stream_not_readable

	;EDI = intermediate buffer
	sub	esp, 512
	mov	edi, esp

.read_loop:
	;size of data to read
	mov	edx, 512
	cmp	edx, ecx
	cmova	edx, ecx

	;read into buffer
	mov	eax, [ebx + STREAM.vtab]
	libcall [eax+STREAM_VTAB.read], [stream], edi, edx
	jc	.error_read

	;return if EOF was reached
	jno	@f
	mov	[ebx+STREAM.rbuf.eof], 1
	add	esp, 512
	sub	ecx, eax
	mov	eax, [count]
	sub	eax, ecx
	jmp	.ro
	@@:

	;loop if we didn't skip enogh
	sub	ecx, eax
	jne	.read_loop
	add	esp, 512

	;we've skipped enough. return EAX=count, OF=0
	mov	eax, [count]

.rno:	and	word [esp], not (1 shl 11) ;clear OF
.rnc:	and	byte [esp], not 1 ;clear CF
.r:	popf
	pop	edi edx ecx ebx
	ret
.rc:	or	byte [esp], 1 ;set CF
	jmp	.r
.ro:	or	word [esp], 1 shl 11 ;set OF
	cmp	[ebx+STREAM.rbuf.eof], 1 ;just some assertion to catch internal errors
	jne	.error_corrupt_stream
	jmp	.rnc

.error_not_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc

.error_stream_not_readable:
	mov	eax, ERR_HANDLE_NOT_READABLE
	jmp	.rc

.error_corrupt_stream:
	mov	eax, ERR_HANDLE_CORRUPT
	jmp	.rc

.error_read:
	;remove buffer from stack
	add	esp, 512
	jmp	.rc
endp