;; @name text.write.format
;; @desc
;;    Writes formatted output to stream. That means, displays format string
;;    with special % values substituted by converted values of additional
;;    arguments.
;; @arg stream
;;    Stream to write to.
;;    For standard streams, you can use {stdout.write} or {stderr.write}.
;; @arg format
;;    Format string. See {output format string} for more information.
;; @arg <arguments>
;;    Remaining arguments depend on format string.
;;    Arguments are pushed on stack normally like with any other procedure.
;;    All arguments are removed from stack by this procedure, no extra code needed.
;; @ret
;;    CF set on error
;; @err ERR_FORMAT_STRING
;;    Incorrect format string. Check {output format string}.
leafproc text.write.format

label .arg1 dword at ebp+16
label .formatstr dword at ebp+12
label .stream  dword at ebp+8
label .convlen dword at ebp-4
label .signed  byte at ebp-5
label .pointed byte at ebp-6
label .size    byte at ebp-7
label .padtype dword at ebp-12
label .padlen  dword at ebp-16
	push	ebp
	mov	ebp, esp
	sub	esp, 16
	push	ebx ecx edx esi edi

	;edi = stream
	mov	edi, [.stream]

	;esi = format string
	mov	esi, [.formatstr]

	;edx = first vararg
	lea	edx, [.arg1]

.format_string_loop:

	;get char
	lodsb

	;check terminating zero
	cmp	al, 0
	je	.done

	;check special char
	cmp	al, '%'
	je	.special_char

	;check escape char (needed for outputting "%")
	cmp	al, '\'
	je	.escaped_char

	jmp	.normal_char

.special_char:

	;set default padding length
	mov	[.padlen], 0		;padding length
	mov	[.padtype], 0		;padding char
	mov	[.pointed], 0		;use pointer
	mov	[.signed], 0		;unsigned by default
	mov	[.size], 4		;size=dword

	;get next character
	lodsb

	;check for 0
	cmp	al, 0
	je	.error_format_string_end

.check_0_padding:

	;check for '0' padding prefix
	cmp	al, '0'
	jne	@f
	mov	[.padtype],2
	lodsb
	@@:

.check_padding_size:

	;check if we are on padding size prefix
	cmp	al, '0'
	jb	.get_pointed
	cmp	al, '9'
	ja	.get_pointed
	dec	esi
	lea	eax, [.convlen]
	libcall extconv.dec_d, esi, -1, eax
	jc	.rc
	mov	[.padlen], eax
	add	esi, [.convlen]
	lodsb
	cmp	al, 0
	je	.error_format_string_end

	;if not padding type set yet, set it to default
	cmp	[.padtype], 0
	jne	@f
	mov	[.padtype], 1
	@@:

.get_pointed:
	cmp	al, 'P'
	jne	@f
	mov	[.pointed], 1
	lodsb
	cmp	al, 0
	je	.error_format_string_end
	@@:

.get_size_override:
	mov	[.size], 1
	cmp	al, 'B'
	je	@f
	mov	[.size], 2
	cmp	al, 'W'
	je	@f
	mov	[.size], 8
	cmp	al, 'Q'
	je	@f
	mov	[.size], 4
	cmp	al, 'D'
	je	@f
	jmp	.get_signed_override
	@@:
	lodsb

.get_signed_override:
	cmp	al, 'i'
	jne	@f
	mov	[.signed], 1
	lodsb
	cmp	al, 0
	je	.error_format_string_end
	@@:

.get_type:

	;check if it's decimal
	cmp	al, 'd'
	je	.print_decimal

	;check if it's hexadecimal
	cmp	al, 'h'
	je	.print_hexadecimal
	cmp	al, 'x'
	je	.print_hexadecimal

	;check if it's string constant
	cmp	al, 's'
	je	.print_string_constant

	;check if it's string
	cmp	al, 'S'
	je	.print_string

	;check if it's char
	cmp	al, 'c'
	je	.print_char

	;unknown type char
	add	edx, 4	;for '%' we are processing now
	jmp	.error_format_string


.print_char:
	;get char
	mov	eax, dword [edx]
	add	edx, 4
	cmp	[.pointed], 0
	je	@f
	mov	al, byte [eax]
	@@:

	;check invalid prefixes
	cmp	[.size], 4
	jne	.error_format_string
	cmp	[.signed], 0
	jne	.error_format_string
	cmp	[.padtype], 0
	jne	.error_format_string
	cmp	[.padlen], 0
	jne	.error_format_string

	;print it
	libcall stream.write.byte, edi, eax
	jc	.rc

	jmp	.format_string_loop

.print_string:

	;get arguments
	mov	ecx, dword [edx+4]  ;string buffer length
	mov	ebx, dword [edx]    ;string pointer
	add	edx, 8

	;check prefixes
	cmp	[.size], 4
	jne	.error_format_string
	cmp	[.pointed], 0
	jne	.error_format_string

	;display string
	libcall text.write.str, edi, ebx, ecx
	jc	.rc

	;get length to pad
	cmp	[.padlen], 0
	je	@f
	libcall str.len, ebx, ecx
	jc	.rc
	sub	[.padlen], eax
	jnc	@f
	mov	[.padlen], 0   ;if string is longer than padding
	@@:

	;display padding
@@:	cmp	[.padlen], 0
	je	@f
	dec	[.padlen]
	libcall stream.write.byte, edi, dword ' '
	jc	.rc
	jmp	@b
	@@:

	jmp	.format_string_loop


.print_string_constant:

	;get arguments
	mov	ebx, dword [edx]    ;string pointer
	add	edx, 4

	;check prefixes
	cmp	[.size], 4
	jne	.error_format_string
	cmp	[.pointed], 0
	jne	.error_format_string

	;display string
	libcall text.write, edi, ebx
	jc	.rc

	;get length to pad
	cmp	[.padlen], 0
	je	@f
	libcall str.len, ebx, ecx
	jc	.rc
	sub	[.padlen], eax
	jnc	@f
	mov	[.padlen], 0   ;if string is longer than padding
	@@:

	;display padding
@@:	cmp	[.padlen], 0
	je	@f
	dec	[.padlen]
	libcall stream.write.byte, edi, dword ' '
	jc	.rc
	jmp	@b
	@@:

	jmp	.format_string_loop

.print_hexadecimal:

	;ECX: EBX = argument
	xor	ecx, ecx
	mov	ebx, [edx]
	add	edx, 4

	;if argument is not pointed, and it's a qword, get also upper dword to ecx
	cmp	[.pointed], 0
	jne	taken @f
	cmp	[.size], 8
	jne	.print_hex.arg_loaded
	mov	ecx, [edx]
	add	edx, 4
	jmp	.print_hex.arg_loaded
	@@:

	;if argument pointer, read proper amount of bytes
	cmp	[.size], 1
	jne	untaken @f
	mov	bl, byte [ebx]
	jmp	.print_hex.arg_loaded
	@@:
	cmp	[.size], 2
	jne	taken @f
	mov	bx, word [ebx]
	jmp	.print_hex.arg_loaded
	@@:
	cmp	[.size], 4
	jne	taken @f
	mov	ebx, dword [ebx]
	jmp	.print_hex.arg_loaded
	@@:
	mov	ecx, dword [ebx+4]
	mov	ebx, dword [ebx]
	.print_hex.arg_loaded:

	;check byte override
	cmp	[.size], 1
	jne	taken @f
	movzx	ebx, bl
	cmp	[.signed], 1
	jne	@f
	movsx	ebx, bl
	@@:

	;check word override
	cmp	[.size], 2
	jne	taken @f
	movzx	ebx, bx
	cmp	[.signed], 1
	jne	@f
	movsx	ebx, bx
	@@:

	;sign-extend dword to qword if needed
	cmp	[.size], 4
	ja	@f
	cmp	[.signed], 1
	jne	@f
	cmp	ebx, 0
	jge	@f
	dec	ecx
	@@:

	;display
	cmp	[.signed], 1
	je	.print_hex_signed
	libcall text.extwrite.q_hex, edi, ebx, ecx, [.padtype], [.padlen]
	jc	.rc
	jmp	@f
.print_hex_signed:
	libcall text.extwrite.iq_hex, edi, ebx, ecx, [.padtype], [.padlen], 0
	jc	.rc
@@:

	jmp	.format_string_loop


.print_decimal:

	;ECX:EBX = argument
	xor	ecx, ecx
	mov	ebx, [edx]
	add	edx, 4

	;if argument is not pointed, and it's a qword, get also upper dword to ecx
	cmp	[.pointed], 0
	jne	taken @f
	cmp	[.size], 8
	jne	.print_dec.arg_loaded
	mov	ecx, [edx]
	add	edx, 4
	jmp	.print_dec.arg_loaded
	@@:

	;if argument pointer, read proper amount of bytes
	cmp	[.size], 1
	jne	untaken @f
	mov	bl, byte [ebx]
	jmp	.print_dec.arg_loaded
	@@:
	cmp	[.size], 2
	jne	taken @f
	mov	bx, word [ebx]
	jmp	.print_dec.arg_loaded
	@@:
	cmp	[.size], 4
	jne	taken @f
	mov	ebx, dword [ebx]
	cmp	ebx, 0
	jmp	.print_dec.arg_loaded
	@@:
	mov	ecx, dword [ebx+4]
	mov	ebx, dword [ebx]
	.print_dec.arg_loaded:

	;check byte
	cmp	[.size], 1
	jne	@f
	movzx	ebx, bl
	cmp	[.signed], 1
	jne	@f
	movsx	ebx, bl
	@@:

	;check word
	cmp	[.size], 2
	jne	@f
	movzx	ebx, bx
	cmp	[.signed], 1
	jne	@f
	movsx	ebx, bx
	@@:

	;sign-extend dword to qword if needed
	cmp	[.size], 4
	ja	@f
	cmp	[.signed], 1
	jne	@f
	cmp	ebx, 0
	jge	@f
	dec	ecx
	@@:

	;display
	cmp	[.signed], 1
	je	.print_dec_signed
	libcall text.extwrite.q_dec, edi, ebx, ecx, [.padtype], [.padlen]
	jc	.rc
	jmp	@f
.print_dec_signed:
	libcall text.extwrite.iq_dec, edi, ebx, ecx, [.padtype], [.padlen], 0
	jc	.rc
@@:

	jmp	.format_string_loop


.escaped_char:

	;escaped char - load next char
	lodsb

	;error if it is end of string
	cmp	al, 0
	je	.error_format_string_end

.normal_char:

	;plain character, just write it
	libcall stream.write.byte, edi, eax
	jc	.rc

.continue_format_string_loop:
	jmp	.format_string_loop

.done:

.rnc:	clc
.r:	lea	eax, [edx-4]		;eax = pointer to last argument
	pop	edi esi edx ecx ebx	;restore regs
	mov	esp, ebp      ;remove local variables
	pop	ebp	      ;restore ebp
	pop	dword [eax]   ;copy return address to place of last argument
	mov	esp, eax      ;move stack pointer there
	retn		      ;and return

.rpop:	pop	dword [edx-8]		;re-save value of eax from stack to 2nd last argument
	lea	eax, [edx-8]		;eax = pointer to 2nd last argument
	pop	edi esi edx ecx ebx	;restore regs
	mov	esp, ebp      ;remove local variables
	pop	ebp	      ;restore ebp
	pop	dword [eax+4] ;copy return adress to place of last argument
	mov	esp, eax      ;move stack pointer to saved EAX
	pop	eax	      ;restore EAX
	retn		      ;and return

	jmp	.r

.rc:
	;we must count non-escaped % chars in format string to skip remaining args
	push	eax  ;save error code

.onerror_skip_nextchar:
	lodsb

	;char 0 = end of string = we can finally exit
	cmp	al, 1
	jb	.rpop	;jb=jc

	;handle %
	cmp	al, '%'
	je	.onerror_percent

	;handle escaped character
	cmp	al, '\'
	jne	.onerror_skip_nextchar
	lodsb
	cmp	al, 0
	je	.error_format_string_onerror
	jmp	.onerror_skip_nextchar

.onerror_percent:
	lodsb

	;error if end-of-string
	cmp	al,0
	je	.error_format_string_onerror

	;skip all numbers after "%"
	cmp	al, '0'
	jb	@f
	cmp	al, '9'
	jbe	.onerror_percent

	;skip 'P'
	cmp	al, 'P'
	jne	@f
	lodsb
	cmp	al, 0
	je	.error_format_string_onerror
	@@:

	;skip 'B'
	cmp	al, 'B'
	jne	@f
	lodsb
	cmp	al, 0
	je	.error_format_string_onerror
	@@:

	;skip 'W'
	cmp	al, 'W'
	jne	@f
	lodsb
	cmp	al, 0
	je	.error_format_string_onerror
	@@:

	;skip 'D'
	cmp	al, 'D'
	jne	@f
	lodsb
	cmp	al, 0
	je	.error_format_string_onerror
	@@:

	;skip 'i'
	cmp	al, 'i'
	jne	@f
	lodsb
	cmp	al, 0
	je	.error_format_string_onerror
	@@:

	;decide based on type
	cmp	al, 'S'
	jne	@f		;treat any other type as dword sized  ;TODO - change error on stack
	add	edx, 4
	@@:
	add	edx, 4
	jmp	.onerror_skip_nextchar


	;error in format string
.error_format_string:
	mov	eax, ERR_FORMAT_STRING
	jmp	.rc

	;error after reaching end of string
.error_format_string_end:
	push	ERR_FORMAT_STRING
	stc
	jmp	.rpop

	;error during parsing format string after error :)
	;only when end of string is reached
.error_format_string_onerror:
	mov	dword [esp], ERR_FORMAT_STRING
	stc
	jmp	.rpop

endlp